/*
 * Copyright (c) 1998
 *      The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the University of
 *      California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Written by Tom Henderson
 * Version 1.0: 5/29/98
 */

#ifdef STP
#include <sys/param.h>
#include <sys/systm.h> 
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/protosw.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/errno.h>
 
#include <net/if.h>
#include <net/route.h>
   
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/in_pcb.h>
#include <netinet/ip_var.h>
#include <netinet/stp.h>
#include <netinet/stp_var.h>

/* 
 * This function is called if STP is holding the BGNAK or ENDAK to give the 
 * listening process time to fill the send buffer.  If there is no data to
 * send when this expires, simply send the BGNAK or ENDAK.
 */
void 
stp_wait_for_data_dummy(void *arg)
{
	struct stpcb *sp;
	struct socket *so;
	
	sp = (struct stpcb *) arg;
	if (sp == NULL)
		return;

	so = sp->s_inpcb->inp_socket;

	if (sp->s_state == STP_WAIT_FOR_DATA) {
		if (so->so_snd.sb_cc) {
			/* 
		 	 * Send BGNAK with data. sndcc_open() will check to 
			 * see if a BGNAK_SD_END can be sent instead.
		 	 */
			stp_sndcc_open(sp, STP_PDU_BGNAK_SD); 
		} else {
			/* NOTE: bgnak_timestamp may now overestimate RTT */
			stp_sndcc_open(sp, STP_PDU_BGNAK);
		}
		sp->s_state = STP_EST;
	} else if (sp->s_state == STP_WAIT_FOR_DATA_CLOSE) {
		if (so->so_snd.sb_cc) {
			if (sp->shutdown_pending || sp->disc_pending) {
				/* Send ENDAK_SD_END */
#ifdef STP_DEBUG
				printf("Sending ENDAK_SD_END\n");
#endif
				stp_sndcc_close(sp, STP_PDU_ENDAK_SD_END);
                        	sp->cc_count = 1;
				stp_start_timer(sp, STP_TIMER_END, 
				    sp->s_srtt ?  ( 2 * (sp->s_srtt >> 
				    SRTT_SHIFT)/MS_PER_TICK) : 
				    INITIAL_SRTT/MS_PER_TICK);
				sp->s_state = STP_LAST_ACK;

			} else {
				/* Send ENDAK_SD */
				stp_sndcc_close(sp, STP_PDU_ENDAK_SD);
				sp->s_state = STP_CLOSE_RECV;
			}
		} else {
			/* 
			 * No data to send.  First, process END PDU as if we
			 * were in STP_EST state, which will trigger ENDAK.  
			 */  
			sp->s_state = STP_EST;
			stp_rcvend(so, sp, sp->rcv_nxt);
			/* 
			 * Next, reissue the DISCONNECT or SHUTDOWN request.
			 * Connection will move from CLOSE_RECV state.
			 */
			if (sp->disc_pending)
				stp_disconnect(sp);
			if (sp->shutdown_pending)
				stp_shutdown(sp);
		}
	} else {
		/* Do nothing */; 
#ifdef STP_DEBUG
		printf("WAIT_FOR_DATA timeout in wrong state %x\n",sp->s_state);
#endif
	}
	return;
}

/*
 * This function is called when a connection control (CC) PDU has been sent 
 * and not responded to (times out).  Resend the PDU, and restart the 
 * timer while exponentially backing it off.
 */
void 
stp_sendcc_dummy(void *arg)
{
	struct stpcb *sp;
	int to_value, error;

	sp = (struct stpcb *) arg;
	
	if (sp == NULL)
		return;

	/* Both BGN and END timers should not be running */
	ASSERT(!((sp->s_timer & STP_TIMER_BGN)&&(sp->s_timer & STP_TIMER_END)));

	if (sp->cc_count) {
		if (sp->cc_count >= MAX_CC_ATTEMPTS) {
			/* p. 843 */
			sp = stp_drop(sp, sp->s_softerror ? 
			    sp->s_softerror : ETIMEDOUT);
			return;
		}
		if (sp->s_srtt)	
			to_value = ((sp->s_srtt >> SRTT_SHIFT) << 
			    sp->cc_count)/MS_PER_TICK;
		else 
			to_value = (INITIAL_SRTT << sp->cc_count)/MS_PER_TICK;
		switch (sp->s_state) {
		case STP_BGN_SENT:
			error = stp_sndcc_open(sp, STP_PDU_BGN);
			stp_start_timer(sp, STP_TIMER_BGN, to_value);
			break;
		case STP_BGN_DATA_SENT:
		case STP_BGN_DATA_END_SENT:
			error = stp_sndcc_open(sp, STP_PDU_BGN_SD);
			stp_start_timer(sp, STP_TIMER_BGN, to_value);
			break;
		case STP_HALF_CLOSE_PENDING:
		case STP_FULL_CLOSE_PENDING:
		case STP_LAST_ACK:
			error = stp_sndcc_close(sp, STP_PDU_END);
			stp_start_timer(sp, STP_TIMER_END, to_value);
			break;
		}
		sp->cc_count++;
	} else {
		/* Do nothing */;
#ifdef STP_DEBUG
		printf("Forgot to cancel this timeout timer %x state %x\n", 
		    sp->s_timer, sp->s_state);
#endif
	}
}

/* 
 * If data transfer is occuring in the forward direction, this function is
 * scheduled and called every system tick (10ms).  Data bursts and poll packets
 * are sent from this function.  If a POLL is scheduled to be sent between
 * now and the next scheduled data burst, it is sent now as a piggybacked POLL.
 */
void 
stp_send_dummy(void *arg)
{
	struct stpcb *sp;
	int temp;

	sp = (struct stpcb *) arg;

	if (sp == NULL)
		return;

	/* Turn off SEND timer */
	sp->s_timer &= ~STP_TIMER_SEND;

	if (1 == sp->send_pending && 1 == sp->poll_pending) {
		/* Both POLL and SD can be sent now */
		if (!sp->poll_preempted)
			stp_output(sp, sp->burst_size, POLL);
		else {
			stp_output(sp, sp->burst_size, NO_POLL);
			sp->poll_preempted = FALSE;
		}
		stp_sched_poll(sp);
	} else if (1 == sp->send_pending) {
		/* 
		 * Only scheduled to send data now.  However, if POLL will 
		 * occur between now and the next scheduled burst, might as 
		 * well POLL now and suppress the regularly scheduled POLL.
	 	 */
		if ((sp->poll_pending <= sp->interval) && !sp->poll_preempted) {
			/* save poll_pending, in case stp_output resets it */
			temp = sp->poll_pending;
			stp_output(sp, sp->burst_size, POLL);
			sp->poll_pending = temp;
			/* poll_preempted will block POLL tx. next time */
			sp->poll_preempted = TRUE;
		} else {
			stp_output(sp, sp->burst_size, NO_POLL);
			if (sp->poll_pending)
				sp->poll_pending--;	
		}
	} else if (1 == sp->poll_pending) {
		if (!sp->poll_preempted)
			stp_sndpoll(sp);
		else {
			sp->poll_preempted = FALSE;
			stp_sched_poll(sp);
		}
		if (sp->send_pending)
			sp->send_pending--;	
	} else {
		/* Decrement countdown variables */
		if (sp->send_pending)
			sp->send_pending--;	
		if (sp->poll_pending)
			sp->poll_pending--;	
	}
	/* Restart timer if needed */
	if (!(sp->s_timer & STP_TIMER_SEND) && 
		(sp->send_pending || sp->poll_pending)) {
		stp_start_timer(sp, STP_TIMER_SEND, 1);
	}
}

/* 
 * This function checks whether there is outstanding data to send.  If 
 * trans_stack == 0, this means that the window may have been exceeded.  If
 * sb_cc - snd_off == 0, this means that all data has been sent.   
 */
void 
stp_sched_output(sp)
struct stpcb *sp;
{
	struct socket *so = sp->s_inpcb->inp_socket;

	/* If no window or no data, set send_pending to 0, to turn off poll */
	if (!sp->trans_stack || ((so->so_snd.sb_cc - sp->snd_off) == 0)) {
		sp->send_pending = 0;
		return;
	}
	/* otherwise, set send_pending to the interval value & start the timer*/
	sp->send_pending = sp->interval;
	if (!(sp->s_timer & STP_TIMER_SEND))
		stp_start_timer(sp, STP_TIMER_SEND, 1);
}


/* 
 * This function should be called whenever events that change the amount of
 * eligible data that can be sent occur.  Examples are:  reception of a STAT,
 * close congestion window, reception of data from user process, reception of
 * BGNAK.
 * 
 * The number of eligible new packets to send is stored in sp->trans_stack.
 * If sp->trans_stack is positive, the variables sp->burst_size and
 * sp->inteval govern the sending behavior every 10ms ("delayed send").   
 * 
 * Two separate policies are coded:  WINDOW_CONTROL and RATE_CONTROL (more...).
 * Rate control has a granularity of 1 pkt/tick, or 1 tick/pkt.  This has most
 * profound effect at high rates (1 pkt/tick is approximately 1.1 Mb/s for
 * a packet size of 1428 bytes).  The integer division below rounds to the 
 * nearest integer: X/Y + 0.5.
 * 
 * NOTE:  If application was blocked from sending new data until space 
 * was freed, this routine may execute before application can fill up the 
 * send buffer again, leading to an underestimation of available data to 
 * send.  Solutions:  Use large send buffers, or call this function when 
 * application submits more data (i.e., modify stp_usrreq.c). 
 */
void
stp_set_output(sp)
    register struct stpcb *sp;
{	
	struct socket *so = sp->s_inpcb->inp_socket;
	int win_pkts, pkts_in_buffer, count, rtt;

	if (sp->s_state < STP_BGN_SENT)
		return;
	if (WINDOW_CONTROL) {
		/* 
 		 * Count number of eligible new transmissions and adjust 
		 * the scheduling variables (interval and burst_length) 
		 * first, determine number of outstanding packets possible.
		 */
		pkts_in_buffer = so->so_snd.sb_cc/sp->s_maxseg + 
			(so->so_snd.sb_cc % sp->s_maxseg ? 1 : 0);
		win_pkts = min(sp->snd_wnd, sp->snd_cwnd/sp->s_maxseg);
		win_pkts = min(win_pkts, pkts_in_buffer);
		/* 
		 * Win_pkts/srtt is the rate at which we should send 
		 * snd_nxt-snd_una is number of packets outstanding.
		 */
		if (win_pkts)
			sp->trans_stack = win_pkts - 
			    seq_subtract(sp->snd_nxt,sp->snd_una);
		else
			sp->trans_stack = 0;
		if (sp->trans_stack < 0)
			sp->trans_stack = 0;

		/* 
	 	 * Get rtt for this connection, but if s_srtt == 0, keep it 0. 
		 * we will space out new packets over the course of srtt 
		 */
		rtt = (sp->s_srtt ? (sp->s_srtt >> SRTT_SHIFT) : INITIAL_SRTT);
		/* 
		 * Calculate burst_size (number to send every tick) and 
		 * interval (number of ticks between bursts). Generally, either 
		 * burst_size or interval will be >= 1, with the other = 1 
	   	 */
		if (sp->trans_stack) {
			/* 
			 * If we want to send N packets over the next K 
			 * milliseconds, we want to send N/K packets/ms.  
			 * BSD clock granularity is 10 ms = * 1 tick, so if 
			 * N/K >= 1, * send (N * 10)/K packets/tick.
			 * If N/K < 1, send a packet every K/(N * 10) ticks.
			 */
			count = (sp->trans_stack * 10) / rtt;
			if (count >= 1) { 
				/* 
				 * will send count packets every tick, bounded 
				 * by MAXBURST
				 */
				sp->burst_size = (count < MAXBURST ? count : 
				    MAXBURST);
				sp->interval = 1;
			} else { 
				/* Will not necessarily send every tick */
				sp->interval = rtt / (sp->trans_stack *10);
				sp->burst_size = 1;
				if (!sp->interval) 
					sp->interval = 1;
			}
			if (!(sp->s_timer & STP_TIMER_SEND)) {
				/* 
			 	 * If our send timer is not currently on, this 
				 * means that we must have been blocked from 
				 * sending due to low window or not enough 
				 * user data.  Schedule a SEND at the next 
				 * tick, and a POLL.
			 	 */
				sp->send_pending = 1;
				stp_start_timer(sp, STP_TIMER_SEND, 1);
				stp_sched_poll(sp);
			} else {
				/* 
				 * Send timer was on.  Reschedule; send at 
				 * next tick 
				 */
				sp->send_pending = 1;
				/* 
				 * What to do about POLLing? 
				 */
				if (!sp->poll_pending) {
#ifdef STP_DEBUG
					printf("Send_pending was found without 
					    poll pending\n");
#endif
					stp_sched_poll(sp);
				}
			} 
		} else 
			sp->send_pending = 0;
	} /* if WINDOW_CONTROL */ 
	if (RATE_CONTROL) {
		long pkts_per_tick, ticks_per_pkt; 
		/* 
		 * If in rate control, rate is specified in bits/s.  Convert 
		 * to packets/tick: 
		 * RATE(b/s)/((8 * sp->s_maxseg) b/packet)/(100 ticks/s). 
		 * First determine how many eligible packets to send.
		 */
		pkts_in_buffer = so->so_snd.sb_cc/sp->s_maxseg + 
			(so->so_snd.sb_cc % sp->s_maxseg ? 1 : 0);
		win_pkts = min(sp->snd_wnd, pkts_in_buffer);
		if (win_pkts)
			sp->trans_stack = win_pkts - 
			    seq_subtract(sp->snd_nxt,sp->snd_una);
		else
			sp->trans_stack = 0;
		if (sp->trans_stack < 0)
			sp->trans_stack = 0;
		if (sp->s_maxseg)
			pkts_per_tick = (RATE + 4 * (long) sp->s_maxseg * 100)/ 
			    (8 * (long) sp->s_maxseg * 100);	
		if (RATE)
			ticks_per_pkt = 
			    (8 * (long) sp->s_maxseg * 100 + RATE/2)/RATE;	
		if (sp->trans_stack && (pkts_per_tick || ticks_per_pkt)) {
			if (ticks_per_pkt) {
				sp->burst_size = 1;
				sp->interval = ticks_per_pkt;
			} else {
				sp->burst_size = pkts_per_tick;
				sp->interval = 1;
			}
			if (!(sp->s_timer & STP_TIMER_SEND)) {
				/* 
			 	 * If send timer is not currently on, this 
				 * means that we must have been blocked from 
				 * sending due to low window or not enough 
				 * user data.  Schedule a SEND at the next 
				 * tick, and a POLL.
			 	 */
				sp->send_pending = 1;
				stp_start_timer(sp, STP_TIMER_SEND, 1);
				stp_sched_poll(sp);
			} 
		} else
			sp->send_pending = 0;
	}
	/* 
	 * if (WINDOW_CONTROL && RATE_CONTROL)
	 * Not yet done.  Should allow a minimum and maximum rate, and allow 
	 * window probing within the specified range.  Should be easy to 
	 * implement within this general rate pacing framework.
	 */
}

/* 
 * This function called to send NEW data.  Retransmissions are handled by 
 * stp_process_list() based on USTAT or STAT reception.  "force" counts the 
 * number of packets to send now.  "poll" stimulates SDP with last packet. 
 * Urgent data is sent in its own special packet "SDP_URG".
 */
int
stp_output(sp, force, poll)
    register struct stpcb *sp;
	int force;
	int poll;
{	
	struct socket *so = sp->s_inpcb->inp_socket;
	struct mbuf *m;
	struct stpiphdr *si;
	long len;
	u_long type_field, timestamp, *dataptr;
	int snd_nxt_mod, hdrlen, error, i;
	flag sending_end, sending_urg;

	if (sp->s_state < STP_BGN_SENT || sp->s_state > STP_CLOSE_RECV)
		return (0);

	/* Adjust force to min(force, sp->trans_stack) */
	force = (force < sp->trans_stack ? force : sp->trans_stack); 
	if (force < 1) { 
		/* nothing to send; don't schedule another right now */
		if (poll)
			stp_sndpoll(sp);
		return (0);
	}

	while (force--) {
		sending_urg = sending_end = FALSE;
		snd_nxt_mod = sp->snd_nxt % MAXBUFFER;
		len = so->so_snd.sb_cc - sp->snd_off;
		if (len > sp->s_maxseg)
			len = sp->s_maxseg;
		if (len < 0) {	
			/* Shouldn't happen */
#ifdef STP_DEBUG
			printf("Length goes negative\n");
#endif
			len = 0;
		}

		if (len == 0) {
			if (poll) 
				stp_sndpoll(sp);
			return (0);
		}
		if (len < sp->s_maxseg) {
			/* 
			 * Silly window avoidance.  If Nagle is disabled, 
			 * proceed.  If Nagle is enabled, return if there are 
			 * outstanding segments that are less than s_maxseg 
			 * and if not in urgent mode. 
			 */
			if (!sp->stp_nodelay && (sp->snd_off % sp->s_maxseg) &&
			    !sp->snd_up) {
				if (poll)	
					stp_sndpoll(sp);
				return (0);
			}
		}
		/* 
		 * Urgent data:  If in urgent mode and the next byte of data 
		 * is the urgent byte, send it as an SDP_URG.  If in urgent 
		 * mode, and if sending a full sized segment would move us 
		 * past the urgent pointer, send only up to the urgent byte
		 */
		if ((sp->snd_up - sp->snd_off) > 0) {
			ASSERT(sp->snd_up > 0);
			if ((sp->snd_up - sp->snd_off) == 1) {
				sending_urg = TRUE;
				len = 1;
			} else if ((len + sp->snd_off) >= sp->snd_up) {
				len = sp->snd_up - sp->snd_off - 1;
			}
		}

		hdrlen = sizeof(struct stpiphdr);		
		/* Set type field */
		type_field = STP_PDU_SD;
		if (sending_urg)
			type_field = STP_PDU_SDP_URG;
		/* 
		 * If this is the last one we are sending in this burst AND
		 * the sending process has closed AND this is the last PDU in
		 * the send buffer, turn off poll, and turn on sending_end.
		 */
		if ( !force && (sp->shutdown_pending || sp->disc_pending) &&
			(so->so_snd.sb_cc == (sp->snd_off + len)) ) {
			poll = FALSE;
			sending_end = TRUE;
		}
		if (!force && !sending_urg) {
			/* 
		 	 * If sending process has closed, use this opportunity 
			 * to piggyback the end if it is the last PDU to send.
		 	 */
			if (sending_end) {
				type_field = STP_PDU_SD_END;
				sp->end_sent = TRUE;
			} else if (poll) 
				type_field = STP_PDU_SDP;
		}
		m = stp_alloc_packet(sp, type_field, sp->snd_nxt); 
		if (NULL == m) {
			error = ENOBUFS;
			goto out;
		}
		si = mtod(m, struct stpiphdr *);

		/* Add timestamp if last one is a SDP, or if urg.data is sent */
		if ((!force && poll) || sending_urg) {
			timestamp = current_time();
			dataptr = (u_long *) (mtod(m, caddr_t) + hdrlen);
			*dataptr = htonl(timestamp);
			m->m_len += 4;
			hdrlen += 4;
		}
		/* 
		 * At this point, m_data points to start of IP packet.  hdrlen
		 * counts the number of bytes after which data can be inserted.
		 */
		if (len <= MHLEN - hdrlen - max_linkhdr) {
			m_copydata(so->so_snd.sb_mb, sp->snd_off, (int) len, 
			    mtod(m, caddr_t)+ hdrlen);	
			m->m_len += len;
		} else {
			m->m_next = m_copy(so->so_snd.sb_mb, sp->snd_off, 
			    (int) len);
			if (m->m_next == 0)
				len = 0;
		}
		/* Save record of transmission */
		sp->trans_buffer[snd_nxt_mod].seqno = sp->snd_nxt;
		sp->trans_buffer[snd_nxt_mod].num_bytes = len;
		sp->trans_buffer[snd_nxt_mod].timesent = current_time();
		sp->trans_buffer[snd_nxt_mod].num_trans = 1; 
		sp->snd_off += len;
#ifdef STP_DEBUG
		if (len > sp->s_maxseg)
			printf("snd_off being corrupted; len:%d\n", len);
#endif

		sp->snd_nxt = SEQ_INC(sp->snd_nxt);

		stp_finish_packet(sp, si, m, hdrlen, len);

		/* mbuf data pointer shouldn't point beyond link layer space */
		error = ip_output(m, sp->s_inpcb->inp_options, 
		    &sp->s_inpcb->inp_route, so->so_options & SO_DONTROUTE, 0);

		/* Decrement transmission stack */
		sp->trans_stack--;

#ifdef STP_STATS
		sp->data_bytes_sent += len;
		sp->total_bytes_sent += len + hdrlen;
		if (type_field = STP_PDU_SDP)
			sp->sdp_sent++;
		sp->pkts_sent++;
#endif

		if (error) {
out:
			if (error == ENOBUFS) {
				stp_quench(sp->s_inpcb, 0);
				/* XXX Potential bug here if return w/o 
				 * checking for a poll? 
				 */
				return (0);
			}
		}	
	} /* while (force) */

	ASSERT(sp->trans_stack >= 0);

	/* schedule next output */
	stp_sched_output(sp);

	return (0);
}

/*  
 *  This routine finds the data corresponding to packet number "seqno"
 *  and retransmits the packet.
 *  Function returns -1 if an error was detected.
 */
int 
stp_retransmit(sp, seqno)
	register struct stpcb *sp;
	u_long seqno;
{   
	struct socket *so = sp->s_inpcb->inp_socket;
	struct mbuf *m;
	struct stpiphdr *si;
	u_long i, seqno_field, type_field;
	int hdrlen, error;
	int i_mod, seqno_mod;
	int offset = 0;
	long len;
     
	seqno_mod = seqno % MAXBUFFER;
	/* Find the offset in the send transmit buffer. */
	for (i = sp->snd_una; seq_lt(i, seqno); i = SEQ_INC(i)) {
		i_mod = i % MAXBUFFER;
		offset += sp->trans_buffer[i_mod].num_bytes;
	}   
	len = sp->trans_buffer[seqno_mod].num_bytes;
	hdrlen = sizeof(struct stpiphdr);
	type_field = STP_PDU_SD;
	/* 
	 * If we previously sent an END_SD, and we have to retransmit it, send
	 * another END_SD.
	 */
	if (sp->end_sent && seqno == SEQ_DEC(sp->snd_nxt))
    		type_field = STP_PDU_SD_END;
	m = stp_alloc_packet(sp, type_field, seqno); 
	if (NULL == m) {    
		error = ENOBUFS;    
		goto out;
	}       
	si = mtod(m, struct stpiphdr *);
    
	if (len <= MHLEN - hdrlen - max_linkhdr) {
		/* sp->snd_off indicates where the sender is w.r.t. snd_nxt */
		m_copydata(so->so_snd.sb_mb, offset, (int) len, 
		    mtod(m, caddr_t)+ hdrlen);  
		m->m_len += len;
	} else {
		m->m_next = m_copy(so->so_snd.sb_mb, offset, (int) len);
		if (m->m_next == 0)
			len = 0;
	}       
	/* Save record of transmission */
	sp->trans_buffer[seqno_mod].timesent = current_time();
	sp->trans_buffer[seqno_mod].num_trans += 1;
    
	/* Upon 3rd transmission of a packet, start backing off retxs */
	/* XXX get rid of magic number */
	if ((sp->trans_buffer[seqno_mod].num_trans > 2) &&
	    (sp->snd_backoff < (sp->trans_buffer[seqno_mod].num_trans - 2)) ) {
		sp->snd_backoff = sp->trans_buffer[seqno_mod].num_trans - 2;
		sp->snd_backoff = (sp->snd_backoff > MAXBACKOFF ? MAXBACKOFF :
		    sp->snd_backoff);
#ifdef STP_DEBUG
		printf("Backing off: sp->snd_backoff %d\n", sp->snd_backoff);
#endif
	}
	if (sp->trans_buffer[seqno_mod].num_trans > STP_REROUTE) 
		/* ask IP to find a new route */
		in_losing(sp->s_inpcb);
	if (sp->trans_buffer[seqno_mod].num_trans > STP_MAXRETRANS) {
		/* XXX Is this the right error value?  Do we want to reset? */
#ifdef STP_DEBUG
		printf("Over max retransmissions %d, stuck on %x\n", 
		    sp->trans_buffer[seqno_mod].num_trans, seqno);
#endif
		return (0);
	}
    
	stp_finish_packet(sp, si, m, hdrlen, len);
    
	/* mbuf data pointer should not be pointing beyond link layer space */
	error = ip_output(m, sp->s_inpcb->inp_options, &sp->s_inpcb->inp_route,
	    so->so_options & SO_DONTROUTE, 0);

#ifdef STP_STATS
	sp->total_bytes_sent += len + hdrlen;
	sp->pkts_sent++;
	sp->retx_sent++;
#endif
    
	if (error) {
out:    
		if (error == ENOBUFS) {
			stp_quench(sp->s_inpcb, 0);
			return (0);
		}
	}
	return (0);
}       


/* 
 * When there is no pcb for a received packet, this function generates a
 * response.  The value for type_field (either ENDAK or RST) is passed in.
 */
void
stp_reset(m, si, type_field)
	struct mbuf *m;
	struct stpiphdr *si;
	u_long type_field;
{
	struct route *ro = 0;
	u_long seqno_field = 0;

	/* 
	 * No need to allocate another packet.  Free any subsequent mbufs,
	 * assign mbuf data pointer to the stpiphdr pointer, overwrite
	 * sequence number field, exchange addresses and ports, and
	 * zero out checksum field
	 */
	m_freem(m->m_next);
	m->m_next = 0;
	m->m_data = (caddr_t) si;
#define xchg(a,b,type) { type t; t=a; a=b; b=t;}
	xchg(si->si_dst.s_addr, si->si_src.s_addr, u_long);
	xchg(si->si_dport, si->si_sport, u_short);
#undef xchg
	seqno_field = (seqno_field | (type_field << 24));
	si->si_seq = htonl(seqno_field);
	si->si_sum = 0;
	/* Leave si->si_inc as it is */
	m->m_len = sizeof (struct stpiphdr);
	m->m_pkthdr.rcvif = (struct ifnet *) 0;
		
	/* stp_finish_packet() can't be called here because there is no pcb */
	si->si_len = htons((u_short) (sizeof(struct stphdr)));
	si->si_sum = stp_cksum(m, sizeof (struct stphdr));
	m->m_pkthdr.len = sizeof(struct stpiphdr);
	((struct ip *) si)->ip_len = m->m_pkthdr.len;
	((struct ip *) si)->ip_ttl = ip_defttl;

	/* mbuf data pointer should not be pointing beyond link layer space */
	(void) ip_output(m, NULL, ro, 0, NULL);
}

/* 
 * This function called to output one of the following PDUs:
 * BGN, BGN_SD, BGNAK, BGNAK_SD.  If BGN_SD or BGNAK_SD is requested, 
 * there is only a small amount of data to send, and a shutdown or disconnect 
 * is pending, then the BGN_SD or BGNAK_SD has an _END appended.
 */
int
stp_sndcc_open(sp, pkttype)
	register struct stpcb *sp;
	u_long pkttype;
{
	struct socket *so = sp->s_inpcb->inp_socket;
	struct mbuf *m;
	struct stpiphdr *si;
	int hdrlen;
	int error, index;
	u_long *dataptr;
	int datalen = 0;	/* amount of user data appended */
	int data_offset = 0;	/* offset from header */
	long len = 0;		/* total length of STP packet */

	u_long timestamp = current_time(); 
	
	ASSERT(pkttype & 0x10);	/* Test for BGN* packet type */
	/* First, determine packet type to send */
	if (pkttype & 0x01) { /* Data included */
		if (so->so_snd.sb_cc) {
			data_offset = 8; /* window and timestamp */
			index = sp->snd_iss % MAXBUFFER;
			if (sp->snd_nxt == sp->snd_iss) 
				datalen = min(sp->s_maxseg - data_offset,
				    so->so_snd.sb_cc);    
			else {
				/* Have sent data before (retransmission) */
				datalen = sp->trans_buffer[index].num_bytes;	
				/*
				 * Check if datalen = length of send buffer
				 * and if there is disconnect or shutdown
				 * pending.  If so, change pkttype to *_END
				 */
				if (datalen == so->so_snd.sb_cc &&
				    (sp->shutdown_pending || sp->disc_pending)){
					if (pkttype == STP_PDU_BGN_SD)
						pkttype = STP_PDU_BGN_SD_END;
					if (pkttype == STP_PDU_BGNAK_SD)
						pkttype = STP_PDU_BGNAK_SD_END;
				}
						
			}
		} else {
			/* No data to send (log protocol error?) */
			if (pkttype == STP_PDU_BGN_SD)
				pkttype = STP_PDU_BGN;
			if (pkttype == STP_PDU_BGNAK_SD)
				pkttype = STP_PDU_BGNAK;
		}
	}
	/* Next, allocate packet and fill in common fields.  */
	m = stp_alloc_packet(sp, pkttype, sp->snd_iss);
	if (NULL == m) {	
		stp_quench(sp->s_inpcb, 0);
		return (0);
	}
	si = mtod(m, struct stpiphdr *);
	hdrlen = sizeof (struct stpiphdr);

	/* Add send window (4 bytes) */
	sp->rcv_wnd = sbspace(&so->so_rcv) / sp->s_peermaxseg;
	dataptr = (u_long *) (mtod(m, caddr_t) + hdrlen);
	*dataptr = htonl(sp->rcv_wnd);
	len  += 4;
	m->m_len += 4;

	/* Add timestamp (4 bytes) */
	dataptr++;
	if (pkttype & 0x02)  /* Some type of BGNAK */
		*dataptr = htonl(sp->bgnak_timestamp);
	else
		*dataptr = htonl(timestamp);
	len += 4;
	m->m_len += 4;

#ifdef STP_STATS
	sp->starttime = timestamp;
#endif
	/* Append data */
	if (data_offset) {
   		if (datalen <= MHLEN - hdrlen - max_linkhdr - data_offset) {
        		m_copydata(so->so_snd.sb_mb, 0, (int) datalen, 
			    mtod(m, caddr_t) + hdrlen + data_offset);
        		m->m_len += datalen;
			len += datalen;
    		} else {
        		m->m_next = m_copy(so->so_snd.sb_mb, 0, (int) datalen);
			len += datalen;
			if (m->m_next == 0)
				len = 0;
    		}
		sp->trans_buffer[index].seqno = sp->snd_iss;
		sp->trans_buffer[index].num_bytes = datalen;
		sp->trans_buffer[index].timesent = current_time();
		sp->trans_buffer[index].timesent += 1;
		if (sp->snd_nxt == sp->snd_iss) {
			sp->snd_nxt = SEQ_INC(sp->snd_nxt);
			sp->snd_off += datalen;
#ifdef STP_DEBUG
			if (len > sp->s_maxseg)
				printf("snd_off corrupted; 
				    datalen:%d\n", datalen);
#endif
		}
	}

	stp_finish_packet(sp, si, m, hdrlen, len);

	/* mbuf data pointer should not be pointing beyond link layer space */
	error = ip_output(m, sp->s_inpcb->inp_options, &sp->s_inpcb->inp_route,
	    so->so_options & SO_DONTROUTE, 0);
#ifdef STP_STATS
	sp->total_bytes_sent += len + hdrlen;
	sp->pkts_sent++;
#endif
	return error;
}

/* 
 * This function called to output one of the following PDUs:
 * END, ENDAK, ENDAK_SD_END, ENDAK_SD.  ENDAK_SD_END and ENDAK_SD are only
 * sent if peer had sent a BGN_SD_END to start the connection.
 */
int
stp_sndcc_close(sp, pkttype)
	register struct stpcb *sp;
	u_long pkttype;
{
	struct socket *so = sp->s_inpcb->inp_socket;
	struct mbuf *m;
	struct stpiphdr *si;
	int hdrlen;
	int error, index;
	u_long *dataptr;
	long len = 0;

	ASSERT(pkttype & 0x20); /* Some type of END */
	/* First, determine packet type to send */
	if (pkttype & 0x01) { /* Data included */
		if (so->so_snd.sb_cc) {
			index = sp->snd_iss % MAXBUFFER;
			if (sp->snd_nxt == sp->snd_iss) 
				len = min(sp->s_maxseg, so->so_snd.sb_cc);    
			else 
				/* Retransmission */
				len = sp->trans_buffer[index].num_bytes;	
		} else {
			/* No data to send (log protocol error?) */
			; /* Do nothing */
		}
	}
	m = stp_alloc_packet(sp, pkttype, sp->snd_nxt);
	if (NULL == m) {	
		stp_quench(sp->s_inpcb, 0);
		return (0);
	}
	si = mtod(m, struct stpiphdr *);
	hdrlen = sizeof (struct stpiphdr);

	/* Append data */
	if (pkttype & 0x01) { 
	   	if (len <= MHLEN - hdrlen - max_linkhdr) {
			/* Will fit into single mbuf */	
       			m_copydata(so->so_snd.sb_mb, 0, 
			    (int) len, mtod(m, caddr_t) + hdrlen);
        		m->m_len += len;
    		} else {
			/* mbuf chain needed */
        		m->m_next = m_copy(so->so_snd.sb_mb, 0, 
			    (int) len);
			if (m->m_next == 0)
				len = 0;
    		}
		/* Record data transmission */
		sp->trans_buffer[index].seqno = sp->snd_iss;
		sp->trans_buffer[index].num_bytes = len;
		sp->trans_buffer[index].timesent = current_time();
		sp->trans_buffer[index].timesent += 1;
		if (sp->snd_nxt == sp->snd_iss) {
			sp->snd_nxt = SEQ_INC(sp->snd_nxt);
			sp->snd_off += len;
#ifdef STP_DEBUG
			if (len > sp->s_maxseg)
				printf("snd_off corrupted; len:%d\n", len);
#endif
		}
	}
	stp_finish_packet(sp, si, m, hdrlen, len);

	/* mbuf data pointer should not be pointing beyond link layer space */
	error = ip_output(m, sp->s_inpcb->inp_options, &sp->s_inpcb->inp_route,
	    so->so_options & SO_DONTROUTE, 0);
#ifdef STP_STATS
	sp->total_bytes_sent += len + hdrlen;
	sp->pkts_sent++;
#endif
	return error;
}

/* Send a RST PDU for this particular connection instance */
int
stp_sndrst(sp)
    register struct stpcb *sp;
{
	struct socket *so = sp->s_inpcb->inp_socket;
	struct mbuf *m;
	struct stpiphdr *si;
	int error;
	long len = 0;

	u_long timestamp = current_time(); 

	m = stp_alloc_packet(sp, STP_PDU_RST, sp->snd_nxt); 
	if (NULL == m) {	
		/* XXX what to do if allocation fails here? */
		stp_quench(sp->s_inpcb, 0);
		return (0);
	}
	si = mtod(m, struct stpiphdr *);
		
	si->si_sum = stp_cksum(m, sizeof (struct stphdr));

	m->m_pkthdr.len = sizeof (struct stpiphdr);
	((struct ip *) si)->ip_len = m->m_pkthdr.len;
	((struct ip *) si)->ip_ttl = sp->s_inpcb->inp_ip.ip_ttl;
	((struct ip *) si)->ip_tos = sp->s_inpcb->inp_ip.ip_tos;

	/* mbuf data pointer should not be pointing beyond link layer space */
	error = ip_output(m, sp->s_inpcb->inp_options, &sp->s_inpcb->inp_route,
	    so->so_options & SO_DONTROUTE, 0);

	return error;
}

/*  
 * Three cases:
 * i) if no data to send, but connection is open, start KEEPALIVE
 * ii) if window == 0, but data to send, PERSIST
 * iii) otherwise follow WINDOW_CONTROL or RATE_CONTROL policy below
 */
void 
stp_sched_poll(sp)
struct stpcb *sp;
{
	struct socket *so = sp->s_inpcb->inp_socket;
	int rtt;
	u_long win_bytes = min(sp->snd_wnd * sp->s_maxseg, sp->snd_cwnd);

	/* if nothing to send, start keepalive timer if not already running */
	if ((sp->snd_nxt == sp->snd_una) && !so->so_snd.sb_cc) {
		if (!(sp->s_timer & STP_TIMER_KEEPALIVE)) {
			stp_killtimers(sp); 
			stp_start_timer(sp, STP_TIMER_KEEPALIVE, NULL);
		}
		sp->poll_pending = 0;
		return;
	}
	/* if window == 0, but data to send, and no urgent data, persist */
	if ((sp->snd_nxt == sp->snd_una) && so->so_snd.sb_cc && 
			!sp->snd_up && !sp->snd_wnd) {
		if (!(sp->s_timer & STP_TIMER_PERSIST)) {
			stp_start_timer(sp, STP_TIMER_PERSIST, NULL);
			return;
		}
	}	
	/* If not in KEEPALIVE or PERSIST, there must be sending job to do */
	rtt = (sp->s_srtt ? (sp->s_srtt >> SRTT_SHIFT) : 
	    INITIAL_SRTT*MS_PER_TICK);
	if (WINDOW_CONTROL) {
		/* 
	  	 * Policy:  
	 	 * i) If window <= (SDP_THRESHOLD * sp->s_maxseg), low window
	 	 * conditions in effect.  If send_pending, POLL with each 
		 * burst.  If no send pending, POLL roughly every RTT.  
	 	 * ii) if send_pending or snd_up,POLL 3 times per RTT, else 
		 * once per RTT
	  	 * iii) if, when POLLing based on RTT, the POLL interval is 
		 * less than the preconfigured MIN_POLL_INTERVAL, increase it 
		 * to MIN_POLL_INTERVAL
	 	 */
		if (win_bytes <= (SDP_THRESHOLD * sp->s_maxseg)) {
			if (sp->send_pending) {
				ASSERT(sp->trans_stack || sp->snd_up);
				sp->poll_pending = sp->send_pending;
			} else {
				sp->poll_pending = (rtt << 
				    sp->snd_backoff)/MS_PER_TICK;
			}
		} else {
			if (sp->send_pending || sp->snd_up) {
				sp->poll_pending = (rtt << 
				    sp->snd_backoff)/(3 * MS_PER_TICK); 
			} else {
				sp->poll_pending = (rtt << 
				    sp->snd_backoff)/MS_PER_TICK; 
			}
			if (sp->poll_pending < MIN_POLL_INTERVAL/MS_PER_TICK)
				sp->poll_pending = 
				    MIN_POLL_INTERVAL/MS_PER_TICK;
		}
	} 
	if (RATE_CONTROL) {
		if (sp->send_pending || sp->snd_up) 
			sp->poll_pending = (rtt << 
			    sp->snd_backoff)/(3 * MS_PER_TICK); 
		else 
			sp->poll_pending = (rtt << sp->snd_backoff)/MS_PER_TICK;
		if (sp->poll_pending < MIN_POLL_INTERVAL/MS_PER_TICK)
			sp->poll_pending = MIN_POLL_INTERVAL/MS_PER_TICK;
	}
	/* Sanity check */
	if (sp->poll_pending < 1)
		sp->poll_pending = 1;
	if (!(sp->s_timer & STP_TIMER_SEND)) {
		stp_start_timer(sp, STP_TIMER_SEND, 1);
	}
}

int
stp_sndpoll(sp)
	register struct stpcb *sp;
{
	struct socket *so = sp->s_inpcb->inp_socket;
	struct mbuf *m;
	struct stpiphdr *si;
	long len;
	int hdrlen;
	u_long type_field;
	u_long timestamp;
	int error;
	u_long *dataptr;

	len = 0;
	timestamp = current_time();
	type_field = STP_PDU_POLL;
	if (sp->snd_up)
		type_field = STP_PDU_POLL_URG;
	hdrlen = sizeof (struct stpiphdr);

	m = stp_alloc_packet(sp, type_field, sp->snd_nxt); 
	if (NULL == m) {	
		/* If allocation fails, put connection in slow start */
		stp_quench(sp->s_inpcb, 0);
		return (0);
	}
	si = mtod(m, struct stpiphdr *);
		
	/* Add timestamp */
	dataptr = (u_long *) (mtod(m, caddr_t) + hdrlen);
	*dataptr = htonl(timestamp);
	len += 4;
	m->m_len += 4;
	
	stp_finish_packet(sp, si, m, hdrlen, len);

	/* mbuf data pointer should not be pointing beyond link layer space */
	error = ip_output(m, sp->s_inpcb->inp_options, &sp->s_inpcb->inp_route,
	    so->so_options & SO_DONTROUTE, 0);

	stp_sched_poll(sp);
#ifdef STP_STATS
	sp->total_bytes_sent += len + hdrlen;
	sp->pkts_sent++;
	sp->poll_sent++;
#endif

	return error;
}
	
int
stp_sndstat(sp, timestamp)
	register struct stpcb *sp;
	u_long timestamp;
{
	struct socket *so = sp->s_inpcb->inp_socket;
	struct mbuf *m;
	struct stpiphdr *si;
	int error;
	long win;
	u_long *dataptr;
	u_long i;
	int i_mod;
	int num_elements = 0;
	long len = 0; 
	int hdrlen = sizeof (struct stpiphdr);

	m = stp_alloc_packet(sp, STP_PDU_STAT, sp->rcv_nxt); 
	if (NULL == m) {    
		/* If allocation fails, put connection in slow start */
		stp_quench(sp->s_inpcb, 0);
		return (0);
	}
	si = mtod(m, struct stpiphdr *);

	/* Add timestamp */
	dataptr = (u_long *) (mtod(m, caddr_t) + hdrlen);
	*dataptr = htonl(timestamp);
	len += 4;
	m->m_len += 4;
	dataptr++;

	sp->rcv_wnd = sbspace(&so->so_rcv) / sp->s_peermaxseg;
	*dataptr = htonl(sp->rcv_wnd);
	len += 4;
	m->m_len += 4;
	
	/* Build stat list (first check if one is needed) */
	/* For now, limit size of STAT to fit in mbuf */
	if (seq_lt(sp->rcv_nxt, sp->rcv_max)) {
		i = sp->rcv_nxt;
		while (seq_lt(i, sp->rcv_max) && num_elements <= MAXSTAT) {
			i_mod = i % MAXBUFFER;
			while (seq_lt(i, sp->rcv_max) && 
			    sp->recv_buffer[i_mod].received) {
				i = SEQ_INC(i);	
				i_mod = i % MAXBUFFER;
			}
			/* Add i to list */
			dataptr++;
			*dataptr = htonl(i);
			len += 4;
			m->m_len += 4;
			num_elements++;
			if (seq_lt(i, sp->rcv_max) && num_elements < MAXSTAT) {
				while (seq_lt(i, sp->rcv_max) && 
				    !sp->recv_buffer[i_mod].received) {
					i = SEQ_INC(i);	
					i_mod = i % MAXBUFFER;
				} /* while */
				/* Add i to list */
				dataptr++;
				*dataptr = htonl(i);
				len += 4;
				m->m_len += 4;
				num_elements++;
			} /* if */
		} /* while (i < sp->rcv_max) */
	} /* if (stat list needed) */

	stp_finish_packet(sp, si, m, hdrlen, len);

	/* mbuf data pointer should not be pointing beyond link layer space */
	error = ip_output(m, sp->s_inpcb->inp_options, &sp->s_inpcb->inp_route,
	    so->so_options & SO_DONTROUTE, 0);
#ifdef STP_STATS
	sp->total_bytes_sent += len + hdrlen;
	sp->pkts_sent++;
	sp->stat_sent++;
#endif

	return error;
}
	
int
stp_sndustat(sp)
	register struct stpcb *sp;
{
	struct socket *so = sp->s_inpcb->inp_socket;
	struct mbuf *m;
	struct stpiphdr *si;
	long len;
	int hdrlen;
	int error;
	u_long *dataptr;
	u_long list1 = sp->ustat_list1;
	u_long list2 = sp->ustat_list2;

	len = 0; 
	hdrlen = sizeof (struct stpiphdr);

	m = stp_alloc_packet(sp, STP_PDU_USTAT, sp->rcv_nxt); 
	if (NULL == m) {	
		/* If allocation fails, put connection in slow start */
		stp_quench(sp->s_inpcb, 0);
		return (0);
	}
	si = mtod(m, struct stpiphdr *);
		
	/* Add list elements */
	dataptr = (u_long *) (mtod(m, caddr_t) + hdrlen);
	*dataptr = htonl(list1);
	len += 4;
	m->m_len += 4;
	dataptr += 1;
	*dataptr = htonl(list2);
	len += 4;
	m->m_len += 4;
	
	stp_finish_packet(sp, si, m, hdrlen, len);

	/* mbuf data pointer should not be pointing beyond link layer space */
	error = ip_output(m, sp->s_inpcb->inp_options, &sp->s_inpcb->inp_route,
	    so->so_options & SO_DONTROUTE, 0);

#ifdef STP_STATS
	sp->total_bytes_sent += len + hdrlen;
	sp->pkts_sent++;
	sp->ustat_sent++;
#endif
	return error;
}

/* 
 * Helper function:  Allocate packet and fill in generic fields. 
 * Return NULL if allocation fails.  Otherwise, return mbuf with STP 
 * generic fields (port numbers, type field, seqno, instance number) filled in
 */
struct mbuf *
stp_alloc_packet(sp, type_field, seqno_field)
	register struct stpcb *sp;
	u_long type_field;
	u_long seqno_field;
{
	struct mbuf *m;
	struct stpiphdr *si;

	MGETHDR(m, M_DONTWAIT, MT_HEADER); /* Allocate mbuf */
	if (NULL == m)
		return m;
	m->m_data += max_linkhdr; /* advance data ptr. beyond ethernet hdr. */	
	m->m_len = sizeof (struct stpiphdr);	/* length of STP/IP header */

	m->m_pkthdr.rcvif = (struct ifnet *) 0;
	si = mtod(m, struct stpiphdr *);
	if (sp->s_template == 0)
		panic("stp_alloc_packet");
	/* Copy template of IP and STP headers into mbuf */
	bcopy((caddr_t) sp->s_template, (caddr_t) si, sizeof(struct stpiphdr));
	/* 
	 * Fill in fields.  The sequence number field contains a 24 bit
	 * `instance' number and a 8 bit `type' field in the highest order bits.
	 */
	seqno_field = (seqno_field | (type_field << 24));
	si->si_seq = htonl(seqno_field);
	si->si_inc = htons(sp->s_instance);

	return m;
}

/* Helper function to finish off filling in some mbuf fields. */
void
stp_finish_packet(sp, si, m, hdrlen, len)
	register struct stpcb *sp;
	struct stpiphdr *si;
	struct mbuf *m;
	int hdrlen;
	long len;
{
	/* Length = STP PDU length; Checksum covers STP PDU only */
	si->si_len = htons((u_short) (hdrlen + len - sizeof (struct ip)));
	si->si_sum = stp_cksum(m, hdrlen + len - sizeof (struct ip));

	m->m_pkthdr.len = hdrlen + len;
	((struct ip *) si)->ip_len = m->m_pkthdr.len;
	((struct ip *) si)->ip_ttl = sp->s_inpcb->inp_ip.ip_ttl;
	((struct ip *) si)->ip_tos = sp->s_inpcb->inp_ip.ip_tos;
}
#endif
